(* archive access *)
(* solyga@gmx.de, 2024-07-08 *)
IMPLEMENTATION MODULE Pak;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT IO, FIO, FIOStd, Str, Lib, Lib2, Lib3, SO, Config, Pak0;

TYPE
  TElmDesc = Pak0.TElmDesc;

VAR
  pak: FIO.File;
  wrt: BOOLEAN;  (* writing allowed *)
  hdr: Pak0.THdrDesc;
  toc: ARRAY TVldIdx OF TElmDesc;
  pos: ARRAY TVldIdx OF TLen;
  err: ARRAY TIdx OF TErrs;
  ers: TErrs;
  cnt: TCnt;
  top: TLen;
  mod: BOOLEAN;  (* toc modified *)


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  (*
  IO.WrStr( Config.data.prg );
  IO.WrStr( ': ' );
  *)
  IO.WrStr( 'Pak.' );
  IO.WrStr( s );
  IO.WrLn();
END Message;

PROCEDURE Warning( r-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'WARNING Pak.' );
  SO.WrStr( b, r );
  IO.WrStr( s );
  IO.WrLn();
END Warning;

PROCEDURE Error( r-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'ERROR Pak.' );
  SO.WrStr( b, r );
  IO.WrStr( s );
  IO.WrLn();
END Error;

PROCEDURE Fatal( r-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'FATAL Pak.' );
  SO.WrStr( b, r );
  Lib3.FatalError( s );
  HALT;
END Fatal;


PROCEDURE DoSetPos( p: TLen );
BEGIN
  FIO.Seek( pak, p );
  IF FIO.GetPos( pak ) # p THEN Fatal( 'DoSetPos: Seek failed' ) END;
END DoSetPos;

PROCEDURE DoRead( VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;
VAR n: CARDINAL;
BEGIN
  n := FIO.RdBin( pak, x, l );
  RETURN n;
END DoRead;

PROCEDURE DoWrite( x-: ARRAY OF BYTE; l: CARDINAL );
BEGIN
  IF ~ wrt THEN Fatal( 'DoWrite: Archive is read-only' ) END;
  FIO.WrBin( pak, x, l );
END DoWrite;


PROCEDURE Cnt(): TCnt;
BEGIN
  RETURN cnt;
END Cnt;

PROCEDURE GetRef( i: TIdx; VAR r: TRef );
BEGIN
  IF i > cnt THEN (* Fatal( 'GetRef: Invalid index' ) *) Lib2.Fill( r, 0 ); RETURN END;
  IF i = 0 THEN Str.Copy( r, '<TOC>' ); RETURN END;
  r := toc[i].ref;
END GetRef;

PROCEDURE Ofs( i: TIdx ): TLen;
BEGIN
  IF i > cnt THEN (* Fatal( 'Ofs: Invalid index' ) *) RETURN 0 END;
  IF i = 0 THEN RETURN hdr.toc.ofs END;
  RETURN toc[i].rng.ofs;
END Ofs;

PROCEDURE Len( i: TIdx ): TLen;
BEGIN
  IF i > cnt THEN (* Fatal( 'Len: Invalid index' ) *) RETURN 0 END;
  IF i = 0 THEN RETURN hdr.toc.len END;
  RETURN toc[i].rng.len;
END Len;

PROCEDURE Err( i: TIdx ): TErrs;
BEGIN
  IF i > cnt THEN (* Fatal( 'Err: Invalid index' ) *) RETURN ersNone END;
  RETURN err[i];
END Err;

PROCEDURE MkErrStr( VAR s: TErrStr; ers: TErrs );
VAR
  e: TErr;
BEGIN
  s := errStr;
  FOR e := MIN( TErr ) TO MAX( TErr ) DO
    IF ~ ( e IN ers ) THEN s[e] := '-' END;
  END;
END MkErrStr;


PROCEDURE Find( r-: TRef ): TIdx;
VAR i: TIdx;
BEGIN
  FOR i := 1 TO cnt DO
    IF toc[i].ref = r THEN RETURN i END;
  END;
  RETURN 0;
END Find;

PROCEDURE Add( r-: TRef; l: TLen ): TIdx;
VAR i: TIdx;
BEGIN
  i := Find( r );
  IF i > 0 THEN Fatal( 'Add: Reference already exists' ) END;
  IF cnt >= cntMax THEN Fatal( 'Add: Too many elements' ) END;
  i := cnt + 1;
  toc[i].ref     := r;
  toc[i].rng.ofs := top;
  toc[i].rng.len := l;
  cnt := i;
  INC( top, l );
  mod := TRUE;
  IF l = 0 THEN Warning( 'Add: Zero file length' ) END;
  RETURN i;
END Add;

PROCEDURE Read( i: TVldIdx; VAR x: ARRAY OF BYTE; n: CARDINAL ): CARDINAL;
VAR l, p: TLen;
BEGIN
  IF i > cnt THEN RETURN 0 END;
  IF n > SIZE( x ) THEN n := SIZE( x ) END;
  l := toc[i].rng.len;
  p := pos[i];
  IF p > l THEN p := l END;
  IF TLen( n ) > l - p THEN n := CARDINAL( l - p ) END;
  (* ( p <= l ) & ( n <= l - p ) *)
  DoSetPos( toc[i].rng.ofs + p );
  n := DoRead( x, n );
  pos[i] := p + TLen( n );
  RETURN n;
END Read;

PROCEDURE Write( i: TVldIdx; x-: ARRAY OF BYTE; n: CARDINAL ): CARDINAL;
VAR l, p: TLen;
BEGIN
  IF ~ wrt THEN Fatal( 'Write: Archive is read-only' ) END;
  IF i > cnt THEN RETURN 0 END;
  IF n > SIZE( x ) THEN n := SIZE( x ) END;
  l := toc[i].rng.len;
  p := pos[i];
  IF p > l THEN p := l END;
  IF TLen( n ) > l - p THEN n := CARDINAL( l - p ) END;
  (* ( p <= l ) & ( n <= l - p ) *)
  DoSetPos( toc[i].rng.ofs + p );
  DoWrite( x, n );
  pos[i] := p + TLen( n );
  RETURN n;
END Write;


PROCEDURE Pos( i: TVldIdx ): TLen;
BEGIN
  IF i > cnt THEN RETURN 0 END;
  RETURN pos[i];
END Pos;

PROCEDURE SetPos( i: TVldIdx; p: TLen );
VAR l: TLen;
BEGIN
  IF i > cnt THEN RETURN END;
  l := toc[i].rng.len;
  IF p > l THEN p := l END;
  pos[i] := p;
END SetPos;


PROCEDURE GetRng( i: TIdx; VAR x: Pak0.TRngDesc );
BEGIN
  IF i > cnt THEN (* Fatal( 'GetRng: Invalid index' ) *) Lib2.Fill( x, 0 ); RETURN END;
  IF i = 0 THEN x := hdr.toc; RETURN END;
  x := toc[i].rng;
END GetRng;

PROCEDURE Contains( x-, y-: Pak0.TRngDesc ): BOOLEAN;
BEGIN
  (* x contains y *)
  RETURN ( x.ofs <= y.ofs ) & ( x.len >= y.len ) & ( ( y.ofs - x.ofs ) <= ( x.len - y.len ) );
END Contains;

PROCEDURE Overlap( x-, y-: Pak0.TRngDesc ): BOOLEAN;
BEGIN
  (* x and y overlap *)
  RETURN ( x.ofs <= y.ofs ) & ( y.ofs - x.ofs < x.len  ) OR
         ( y.ofs <= x.ofs ) & ( x.ofs - y.ofs < y.len  );
END Overlap;

PROCEDURE UpdErrs();
VAR
  d, ri, rj: Pak0.TRngDesc;
  i, j: TIdx;

  PROCEDURE ERng();
  BEGIN
    INCL( err[i], eRng );
    INCL( ers, eRng );
  END ERng;

  PROCEDURE EOvl();
  BEGIN
    INCL( err[i], eOvl );
    INCL( err[j], eOvl );
    INCL( ers, eOvl );
  END EOvl;

BEGIN
  Lib2.Fill( err, 0 );
  ers := ersNone;
  d.ofs := SIZE( hdr );
  d.len := FIO.Size( pak ) - SIZE( hdr );
  FOR i := 0 TO cnt DO
    GetRng( i, ri );
    IF ~ Contains( d, ri ) THEN ERng() END;
    FOR j := i + 1 TO cnt DO
      GetRng( j, rj );
      IF Overlap( ri, rj ) THEN EOvl() END;
    END;
  END;
END UpdErrs;

PROCEDURE UpdTop();
VAR
  i: TVldIdx;
  r: Pak0.TRngDesc;
  t: TLen;
BEGIN
  top := SIZE( hdr );
  IF ( eRng IN ers ) OR ( cnt = 0 ) THEN RETURN END;
  FOR i := 1 TO cnt DO
    GetRng( i, r );
    t := r.ofs + r.len;
    IF top < t THEN top := t END;
  END;
END UpdTop;


PROCEDURE Load( n-: ARRAY OF CHAR );
VAR
  len: TLen;

  PROCEDURE Chr( c: CHAR ): CHAR;
  BEGIN
    IF ( c < ' ' ) OR ( '~' < c ) THEN c := '.' END;
    RETURN c;
  END Chr;

  PROCEDURE Show1();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Load: "' );
    SO.WrStr( b, n );
    SO.WrStr( b, '" len ' );
    b.c.p := '0';
    SO.WrLngHex( b, len, 8 );
    Message( s );
  END Show1;

  PROCEDURE Show2();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    i: CARDINAL;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Load: id "' );
    FOR i := 0 TO 3 DO
      SO.WrChar( b, Chr(hdr.id[i]) );
    END;
    SO.WrStr( b, '" ofs ' );
    b.c.p := '0';
    SO.WrLngHex( b, hdr.toc.ofs, 8 );
    SO.WrStr( b, ' len ' );
    SO.WrLngHex( b, hdr.toc.len, 8 );
    b.c.p := ' ';
    SO.WrStr( b, ' cnt ' );
    SO.WrLngCard( b, cnt, 1 );
    Message( s );
  END Show2;

  PROCEDURE Show3();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    e: TErrStr;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    MkErrStr( e, ers );
    SO.Init( b, s );
    SO.WrStr( b, 'Load: ers ' );
    SO.WrStr( b, e );
    SO.WrStr( b, ' top ' );
    b.c.p := '0';
    SO.WrLngHex( b, top, 8 );
    Message( s );
  END Show3;

BEGIN
  (* open *)
  IF ~ FIO.Exists( n ) THEN Fatal( 'Load: Archive does not exist' ) END;
  pak := FIO.Open( n );
  len := FIO.Size( pak );
  Show1();
  (* load header *)
  IF DoRead( hdr, SIZE( hdr ) ) # SIZE( hdr ) THEN Fatal( 'Load: Cannot read header' ) END;
  cnt := hdr.toc.len DIV SIZE( TElmDesc );
  Show2();
  IF hdr.id # Pak0.idPak                  THEN Fatal( 'Load: Invalid file ID' ) END;
  IF hdr.toc.ofs < SIZE( hdr )            THEN Fatal( 'Load: Invalid TOC offset (too small)' ) END;
  IF hdr.toc.ofs > len                    THEN Fatal( 'Load: Invalid TOC offset (too large)' ) END;
  IF hdr.toc.len > len - hdr.toc.ofs      THEN Fatal( 'Load: Invalid TOC length (too large for this archive)' ) END;
  IF hdr.toc.len > SIZE( toc )            THEN Fatal( 'Load: Invalid TOC length (too large for this program)' ) END;
  IF hdr.toc.len MOD SIZE( TElmDesc ) # 0 THEN Fatal( 'Load: Invalid TOC length (not multiple of 40H)' ) END;
  (* load toc *)
  DoSetPos( hdr.toc.ofs );
  IF DoRead( toc, hdr.toc.len ) # hdr.toc.len THEN Fatal( 'Load: Cannot read TOC' ) END;
  UpdErrs();
  UpdTop();
  Show3();
END Load;

PROCEDURE Create( n-: ARRAY OF CHAR );
VAR
  len: TLen;

  PROCEDURE Show();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Create: "' );
    SO.WrStr( b, n );
    SO.WrStr( b, '" len ' );
    b.c.p := '0';
    SO.WrLngHex( b, len, 8 );
    b.c.p := ' ';
    SO.WrStr( b, ' cnt ' );
    SO.WrLngCard( b, cnt, 1 );
    Message( s );
  END Show;

BEGIN
  IF ~ wrt THEN Fatal( 'Create: Archive is read-only' ) END;
  pak := FIO.Create( n );
  hdr.id := Pak0.idPak;
  hdr.toc.ofs := SIZE( hdr );
  hdr.toc.len := 0;
  cnt := 0;
  top := SIZE( hdr );
  DoWrite( hdr, SIZE( hdr ) );
  len := SIZE( hdr );
  UpdErrs();
  UpdTop();
  Show();
END Create;


PROCEDURE Show();
CONST
  ena = FALSE;
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
  i, is: TVldIdx;
BEGIN
  IF ~ ena THEN RETURN END;
  IF cnt = 0 THEN RETURN END;
  FOR is := 1 TO cnt DO
    i := is; (* OfsIdx( is ); *)
    SO.Init( b, s );
    SO.WrCard( b, is, 3 );
    SO.WrStr( b, ' ' );
    SO.WrCard( b, i, 3 );
    SO.WrStr( b, ' ' );
    SO.WrLngHex( b, toc[i].rng.ofs, 8 );
    SO.WrStr( b, ' ' );
    SO.WrLngHex( b, toc[i].rng.len, 8 );
    SO.WrStr( b, ' ' );
    SO.WrStr( b, toc[i].ref );
    IO.WrStr( s );
    IO.WrLn();
  END;
END Show;

VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ Config.data.vrb THEN RETURN END;
  SO.Init( b, s );
  SO.WrStr( b, 'DoneModule' );
  Message( s );
  oldTerm();
END DoneModule;

BEGIN
  pak := FIOStd.inv;
  wrt := FALSE;
  Lib2.Fill( hdr, 0 );
  Lib2.Fill( toc, 0 );
  Lib2.Fill( pos, 0 );
  Lib2.Fill( err, 0 );
  ers := ersNone;
  cnt := 0;
  top := 0;
  mod := FALSE;

  CASE Config.data.act OF
  | Config.actCheck:
      Load( Config.data.arc );
  | Config.actList :
      Load( Config.data.arc );
  | Config.actTable:
      Load( Config.data.arc );
(*
  | Config.actCreate:
      wrt := TRUE;
      Create( Config.data.arc );
*)
  ELSE
      Load( Config.data.arc );
  END;
  Show();

  Lib3.Terminate( DoneModule, oldTerm );
END Pak.
